#include "stdafx.h"
#include "Jumpstarter.h"
#include "Loggable.h"

#if JUMPSTART


#define DATABASE_FILE_NAME "Jumpstart.db"
#define MAX_NAME_LEN_FOR_HASH 100

#define NUM_CHUNKS_TO_SHOW 2

#define JS_INVISIBLE 0
#define JS_VISIBLE 1
#define JS_COMPLETED 0xFF
#define JS_MAX_USERS 0xFE

// static members
DbEnv*	CJumpstarter::pDbEnv = NULL;
Db*		CJumpstarter::pDbJumpstart = NULL;
Db*		CJumpstarter::pDbUserBlocks = NULL;
Db*		CJumpstarter::pDbUserChunks = NULL;
Db*		CJumpstarter::pDbHashNames = NULL;

// useful structures
struct StartAndSize {
    uint32 start;
    uint32 size;
};

struct FileUser {
    uchar filehash[16];
    uchar userhash[16];
};

struct FileUserChunk {
    uchar filehash[16];
    uchar userhash[16];
    uint32 chunk_no;
};

// sorting function for StartAndSize records
int compare_start_size(DB *pDb, const DBT *pDbt1, const DBT *pDbt2) {
    StartAndSize* ss1= (StartAndSize*) pDbt1->data;
    StartAndSize* ss2= (StartAndSize*) pDbt2->data;
    if(ss1->start < ss2->start)
        return -1;
    if(ss1->start > ss2->start)
        return 1;
    // start is equal
    if(ss1->size < ss2->size)
        return -1;
    if(ss1->size > ss2->size)
        return 1;
    // equal
    return 0;
}


CJumpstarter::CJumpstarter(CKnownFile* file) {
    kfile = file;
    nParts = kfile->GetPartCount();
}

CJumpstarter::~CJumpstarter(void) {
}

void CJumpstarter::OpenDatabases(DbEnv*	pDbEnvironment){
    pDbEnv = pDbEnvironment;
	pDbJumpstart = new Db(pDbEnv, 0);
	pDbJumpstart->set_pagesize(16 * 1024); // 16Kb page size
	pDbJumpstart->open(NULL, DATABASE_FILE_NAME, "File-Chunks", DB_HASH, DB_CREATE | DB_THREAD|
                DB_AUTO_COMMIT, 0);

	pDbUserBlocks = new Db(pDbEnv, 0);
    pDbUserBlocks->set_flags(DB_DUPSORT);
    pDbUserBlocks->set_dup_compare(&compare_start_size);
	pDbUserBlocks->set_pagesize(16 * 1024); // 16Kb page size
	pDbUserBlocks->open(NULL, DATABASE_FILE_NAME, "FileUserChunk-Blocks", DB_HASH, DB_CREATE | DB_THREAD| 
                DB_AUTO_COMMIT, 0);

    pDbUserChunks = new Db(pDbEnv, 0);
	pDbUserChunks->set_pagesize(16 * 1024); // 16Kb page size
	pDbUserChunks->open(NULL, DATABASE_FILE_NAME, "FileUser-Chunk", DB_HASH, DB_CREATE | DB_THREAD| 
                DB_AUTO_COMMIT, 0);

    pDbHashNames = new Db(pDbEnv, 0);
	pDbHashNames->set_pagesize(16 * 1024); // 16Kb page size
	pDbHashNames->open(NULL, DATABASE_FILE_NAME, "Hash-Name", DB_HASH, DB_CREATE | DB_THREAD| 
                DB_AUTO_COMMIT, 0);
}


void CJumpstarter::CloseDatabases(void) {
    if(pDbEnv == NULL) return;
    pDbHashNames->close(0);
    pDbUserChunks->close(0);
    pDbUserBlocks->close(0);
    pDbJumpstart->close(0);
    pDbEnv = NULL;
}

bool CJumpstarter::ShouldBeEnabledForFile(CKnownFile* file) {
    try {
        uchar* filehash = file->GetFileHash();
        Dbt key(filehash, 16);
        uint32 nChunks = file->GetPartCount();
        uchar* chunks_info = new uchar[nChunks];
        Dbt data(chunks_info,sizeof(uchar)*nChunks);
        data.set_ulen(sizeof(uchar)*nChunks);
        data.set_flags(DB_DBT_USERMEM);

        int res = pDbJumpstart->get(NULL, &key, &data, 0);

        bool all_complete=true;
        if(res != DB_NOTFOUND)
            for(uint32 i=0; i<nChunks; i++) 
                if(chunks_info[i] != JS_COMPLETED) {
                    all_complete = false;
                    break;
                }

        delete[] chunks_info;

        if(res == DB_NOTFOUND)
            return false;
        else {
            if(all_complete) {
                CLoggable::AddDebugLogLine(false,"Jumpstart: file %s has already finished the jumpstart period",file->GetFileName());
                return false;
                }
            else
                return true;
        }
    }
	catch(DbException &dbe)
	{
		AfxMessageBox(dbe.what());
	}
    return false;
}

void CJumpstarter::EnableForFile(CKnownFile* file) {
    if(!ShouldBeEnabledForFile(file)) {
        uchar* filehash = file->GetFileHash();
        uint32 nChunks = file->GetPartCount();
        uchar* chunks_info = new uchar[nChunks];
        for(uint32 i=0; i<nChunks; i++)
            chunks_info[i] = 0;

        Dbt key(filehash, 16);
        Dbt data(chunks_info,sizeof(uchar)*nChunks);

        pDbJumpstart->put(NULL, &key, &data, DB_AUTO_COMMIT);

        AddHashName(NULL, filehash, file->GetFileName() );

        delete[] chunks_info;
    }
}

void CJumpstarter::AddSentBlock(CUpDownClient* client, uint32 start_offset, uint32 togo) {
    if(togo == 0) return;
    if(start_offset/PARTSIZE != (start_offset + togo - 1)/PARTSIZE) {
        // split on part border and do two recursive calls to ourselves
        uint32 next_part_start = ((start_offset/PARTSIZE)+1)*PARTSIZE;
        AddSentBlock(client, start_offset, next_part_start-start_offset);
        AddSentBlock(client, next_part_start, togo - (next_part_start-start_offset));
        return;
    }

    int dbresult;
    int dbresult2;

    // the block belongs to one part here
    uint32 partno = start_offset/PARTSIZE;

    DbTxn *tid;
    pDbEnv->txn_begin(NULL, &tid, 0); // start the transaction

    // add the new record (database is sorted)
    FileUserChunk fuchunk;
    memcpy(&fuchunk.filehash, kfile->GetFileHash(), 16);
    memcpy(&fuchunk.userhash, client->GetUserHash(), 16);
    fuchunk.chunk_no = partno;
    Dbt key_fuchunk(&fuchunk, sizeof(fuchunk));

    StartAndSize start_and_size;
    start_and_size.start = start_offset;
    start_and_size.size = togo;
    Dbt data_start_and_size(&start_and_size, sizeof(start_and_size));

    pDbUserBlocks->put(tid, &key_fuchunk, &data_start_and_size,0);

    // now, do a pass removing overlaps
    int blocks_count;
    bool done=false;
    while(!done) {
        Dbc *dbcp;
        pDbUserBlocks->cursor(tid, &dbcp, 0);
        Dbc *dbcp2;
        pDbUserBlocks->cursor(tid, &dbcp2, 0);

        data_start_and_size.set_flags(DB_DBT_USERMEM);
        data_start_and_size.set_ulen(sizeof(start_and_size));


        StartAndSize start_and_size2;
        Dbt data_start_and_size2(&start_and_size2, sizeof(start_and_size2));
        data_start_and_size2.set_flags(DB_DBT_USERMEM);
        data_start_and_size2.set_ulen(sizeof(start_and_size2));

        dbresult = dbcp->get(&key_fuchunk, &data_start_and_size, DB_SET);
        ASSERT(dbresult==0);
        blocks_count = 1;

        dbresult2 = dbcp2->get(&key_fuchunk, &data_start_and_size2, DB_SET);
        ASSERT(dbresult2==0);
        dbresult2 = dbcp2->get(&key_fuchunk,&data_start_and_size2,DB_NEXT_DUP);
        
        bool need_to_merge=false;
        while(dbresult2==0 && !need_to_merge) {
            // check for overlap
            uint32 new_start;
            uint32 new_size;
            if(start_and_size2.start <= start_and_size.start +start_and_size.size) {
                need_to_merge = true;
                new_start = start_and_size.start;
                new_size = max(start_and_size.start + start_and_size.size, 
                                start_and_size2.start + start_and_size2.size) - new_start;
            }
            //
            if(need_to_merge) {
                dbresult = dbcp->del(0);
                dbresult2 = dbcp2->del(0);
                start_and_size.start=new_start;
                start_and_size.size=new_size;
                pDbUserBlocks->put(tid, &key_fuchunk, &data_start_and_size,0);
            } else {
                dbresult = dbcp->get(&key_fuchunk, &data_start_and_size, DB_NEXT_DUP);
                ASSERT(dbresult==0);
                blocks_count++;
                dbresult2 = dbcp2->get(&key_fuchunk,&data_start_and_size2,DB_NEXT_DUP);
            }
        }
        dbcp->close();
        dbcp2->close();
        if(!need_to_merge && dbresult2 == DB_NOTFOUND)
            done = true;
    } // while(!done)

    if(blocks_count==1) {
        // check for full part
        uint32 partsize;
        if(partno==nParts-1) {
            partsize = kfile->GetFileSize() % PARTSIZE;
        } else
            partsize = PARTSIZE;

        if(start_and_size.start==partno*PARTSIZE && start_and_size.size==partsize) {
            // Yes, this is full part
            CLoggable::AddDebugLogLine(false,"Jumpstart: User %s completed chunk %d of file %s",client->GetUserName(),partno,kfile->GetFileName());

            FileUser fu;
            memcpy(&fu.filehash, kfile->GetFileHash(), 16);
            memcpy(&fu.userhash, client->GetUserHash(), 16);
            Dbt key_fu(&fu, sizeof(fu));

            uchar* chunks_info = new uchar[nParts];
            Dbt data_chunks_info(chunks_info,sizeof(uchar)*nParts);
            data_chunks_info.set_ulen(sizeof(uchar)*nParts);
            data_chunks_info.set_flags(DB_DBT_USERMEM);

            int res = pDbUserChunks->get(tid, &key_fu, &data_chunks_info, 0);
            if(res == DB_NOTFOUND) {
                // the chunk data for this user does not exist,
                // probably he got on queue before we've enabled
                // the Jumpstart on this file. easiest way is
                // to ignore him :)
                CLoggable::AddDebugLogLine(false,"Jumpstart: User %s (file %s) has no chunks info",client->GetUserName(),kfile->GetFileName());
            } else {
                chunks_info[partno] = JS_COMPLETED;
                pDbUserChunks->put(tid, &key_fu, &data_chunks_info, 0);
            }

            Dbt key_filehash(&fu.filehash, sizeof(fu.filehash));
            res = pDbJumpstart->get(tid, &key_filehash, &data_chunks_info, 0);
            if(res == DB_NOTFOUND) {
                // the chunk data for this file does not exist
                // this is a serious error
                CLoggable::AddDebugLogLine(false,"Jumpstart: File %s has no chunks info, disabled !",kfile->GetFileName());
                kfile->SetJumpstartEnabled(false);
            } else {
                chunks_info[partno] = JS_COMPLETED;
                pDbJumpstart->put(tid, &key_filehash, &data_chunks_info, 0);
                // check if the file is completed
                bool has_more_chunks=false;
                for(uint32 i=0; i<nParts; i++) {
                    if(chunks_info[i]!=JS_COMPLETED) {
                        has_more_chunks = true;
                        break;
                    }
                }
                if(!has_more_chunks) {
                    // hooray! the file's done
                    CLoggable::AddDebugLogLine(false,"Jumpstart: File %s has all chunks completed, switching to normal!",kfile->GetFileName());
                    kfile->SetJumpstartEnabled(false);
                }
            }

            delete[] chunks_info;
        } // end if got full part
        
    } // end if(blocks_count==1) - one block in this part after merging

    tid->commit(0);
}

void CJumpstarter::WriteJumpstartPartStatus(CUpDownClient* client, CMemFile* file) {
    DbTxn *tid;
    pDbEnv->txn_begin(NULL, &tid, 0); // start the transaction

    // load chunk data for this user
    FileUser fu;
    memcpy(&fu.filehash, kfile->GetFileHash(), 16);
    memcpy(&fu.userhash, client->GetUserHash(), 16);
    Dbt key_fu(&fu, sizeof(fu));

    uchar* chunks_info = new uchar[nParts];
    Dbt data_chunks_info(chunks_info,sizeof(uchar)*nParts);
    data_chunks_info.set_ulen(sizeof(uchar)*nParts);
    data_chunks_info.set_flags(DB_DBT_USERMEM);

    int res = pDbUserChunks->get(tid, &key_fu, &data_chunks_info, 0);
    if(res == DB_NOTFOUND) {
        // the chunk data for this user does not exist
        //CLoggable::AddDebugLogLine(false,"Jumpstart: New user %s for file %s",client->GetUserName(),kfile->GetFileName());
        AddHashName(tid, client->GetUserHash(), client->GetUserName());
        for(uint32 i=0; i<nParts; i++)
            chunks_info[i]=JS_INVISIBLE;
    }

    // load file chunks
    Dbt key_filehash(&fu.filehash, sizeof(fu.filehash));
    uchar* file_chunks_info = new uchar[nParts];
    Dbt data_file_chunks_info(file_chunks_info,sizeof(uchar)*nParts);
    data_file_chunks_info.set_ulen(sizeof(uchar)*nParts);
    data_file_chunks_info.set_flags(DB_DBT_USERMEM);

    res = pDbJumpstart->get(tid, &key_filehash, &data_file_chunks_info, 0);
    if(res == DB_NOTFOUND) {
        // the chunk data for this file does not exist
        // this is a serious error
        CLoggable::AddDebugLogLine(false,"Jumpstart: File %s has no chunks info in WriteJumpstartPartStatus, disabled!",kfile->GetFileName());
        kfile->SetJumpstartEnabled(false);
    }

    // find chunks to show
    while(NumVisibleInChunksMap(chunks_info) < NUM_CHUNKS_TO_SHOW) {
        uint32 best_part=0xFFFF;
        uint32 best_part_value=0;
        for(uint32 i=0; i<nParts; i++) {
            if( (chunks_info[i] == JS_INVISIBLE) && (file_chunks_info[i] != JS_COMPLETED)) {
                if(best_part==0xFFFF) {
                    // first suitable, remember it
                    best_part = i;
                    best_part_value = file_chunks_info[i];
                } else {
                    // compare parts
                    if(file_chunks_info[i] < best_part_value) {
                        // new one is better
                        best_part = i;
                        best_part_value = file_chunks_info[i];
                    }
                }
            }
        } // end for(uint32 i=0; i<nParts ...
        if(best_part==0xFFFF) {
            // nothing found, show all
            for(uint32 i=0; i<nParts; i++)
                if(chunks_info[i] == JS_INVISIBLE)
                    chunks_info[i] = JS_VISIBLE;
            break; // end iterations
        } else {
            CLoggable::AddDebugLogLine(false,"Jumpstart: Exposing chunk %d of file %s to user %s",best_part, 
                            kfile->GetFileName(),client->GetUserName());
            chunks_info[best_part] = JS_VISIBLE;
            if(file_chunks_info[best_part] < JS_MAX_USERS)
                file_chunks_info[best_part] += 1;
        }
    } // end while(NumVisibleInChunksMap(chunks_info) < NUM_CHUNKS_TO_SHOW)

    // save new data for user and file
    pDbUserChunks->put(tid, &key_fu, &data_chunks_info, 0);
    pDbJumpstart->put(tid, &key_filehash, &data_file_chunks_info, 0);
    
    tid->commit(0);

    // now actualy write the part status
    uint16 parts = nParts;
	file->Write(&parts,2);
	uint16 done = 0;
	while (done != parts){
		uint8 towrite = 0;
		for (uint32 i = 0;i != 8;i++){
			if (chunks_info[done] != JS_INVISIBLE)
				towrite |= (1<<i);

			done++;
			if (done == parts)
				break;
		}
		file->Write(&towrite,1);
	}

    delete[] chunks_info;
    delete[] file_chunks_info;
}


uint32 CJumpstarter::NumVisibleInChunksMap(uchar* chunks_map) {
    uint32 count_visible=0;
    for(uint32 i=0; i<nParts; i++)
        if(chunks_map[i] == JS_VISIBLE)
            count_visible++;
    return count_visible;
}

void CJumpstarter::AddHashName(DbTxn* tid, uchar* hash, char* name){
    Dbt key(hash,16);
    char buffer[MAX_NAME_LEN_FOR_HASH];
    strncpy(buffer, name, MAX_NAME_LEN_FOR_HASH-1);
    buffer[MAX_NAME_LEN_FOR_HASH-1] = 0;
    Dbt data(buffer, strlen(buffer)+1);

    int flags = 0;
    if(tid==NULL)
        flags = DB_AUTO_COMMIT;

    pDbHashNames->put(tid, &key, &data, flags);
}

#endif // JUMPSTART
